'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

exports.GomokuDirectionType = void 0;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(exports.GomokuDirectionType || (exports.GomokuDirectionType = {}));
const gomokuDirections = Array.from(Object.entries({
    [exports.GomokuDirectionType.TOP]: [-1, 0],
    [exports.GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [exports.GomokuDirectionType.RIGHT]: [0, 1],
    [exports.GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [exports.GomokuDirectionType.BOTTOM]: [1, 0],
    [exports.GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [exports.GomokuDirectionType.LEFT]: [0, -1],
    [exports.GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const leftHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.LEFT,
    exports.GomokuDirectionType.TOP_LEFT,
    exports.GomokuDirectionType.TOP,
    exports.GomokuDirectionType.TOP_RIGHT,
];
const rightHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.RIGHT,
    exports.GomokuDirectionType.BOTTOM_RIGHT,
    exports.GomokuDirectionType.BOTTOM,
    exports.GomokuDirectionType.BOTTOM_LEFT,
];
const gomokuDirectionTypes = leftHalfGomokuDirectionTypes.concat(rightHalfGomokuDirectionTypes);

class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_INLINE;
    NEXT_MOVER_BUFFER_FAC;
    TOTAL_POS;
    TOTAL_PLAYERS;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE, NEXT_MOVER_BUFFER_FAC = 0.4) {
        this.MAX_ROW = MAX_ROW;
        this.MAX_COL = MAX_COL;
        this.MAX_INLINE = MAX_INLINE;
        this.TOTAL_POS = MAX_ROW * MAX_COL;
        this.TOTAL_PLAYERS = 2;
        this.NEXT_MOVER_BUFFER_FAC = Math.max(0.1, Math.min(0.9, NEXT_MOVER_BUFFER_FAC));
    }
    idx(r, c) {
        return r * this.MAX_ROW + c;
    }
    idxIfValid(r, c) {
        return this.isValidPos(r, c) ? r * this.MAX_ROW + c : -1;
    }
    reIdx(id) {
        const c = id % this.MAX_ROW;
        const r = Math.round((id - c) / this.MAX_ROW);
        return [r, c];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    isInvalidPos(r, c) {
        return r < 0 || r >= this.MAX_ROW || c < 0 || c >= this.MAX_COL;
    }
    move(r, c, dirType, step) {
        const [dr, dc] = gomokuDirections[dirType];
        const r2 = r + dr * step;
        const c2 = c + dc * step;
        return [r2, c2];
    }
    *validNeighbors(r, c) {
        for (const dirType of gomokuDirectionTypes) {
            const [r2, c2] = this.move(r, c, dirType, 1);
            if (this.isValidPos(r2, c2))
                yield [r2, c2, dirType];
        }
    }
    traverseAllDirections(handle) {
        this.traverseLeftDirections(handle);
        this.traverseRightDirections(handle);
    }
    traverseLeftDirections(handle) {
        const { MAX_ROW, MAX_COL } = this;
        for (const dirType of leftHalfGomokuDirectionTypes) {
            for (let r = 0; r < MAX_ROW; ++r) {
                for (let c = 0; c < MAX_COL; ++c)
                    handle(r, c, dirType);
            }
        }
    }
    traverseRightDirections(handle) {
        const { MAX_ROW, MAX_COL } = this;
        const R = MAX_ROW - 1;
        const C = MAX_COL - 1;
        for (const dirType of rightHalfGomokuDirectionTypes) {
            for (let r = R; r >= 0; --r) {
                for (let c = C; c >= 0; --c)
                    handle(r, c, dirType);
            }
        }
    }
}

class GomokuCountMap {
    context;
    board;
    dirCountMap;
    conShapeCountMap;
    gapShapeCountMap;
    scoreMap;
    constructor(context, board, scoreMap) {
        this.context = context;
        this.board = board;
        this.dirCountMap = new Array(gomokuDirections.length)
            .fill([])
            .map(() => new Uint32Array(context.TOTAL_POS));
        this.conShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.gapShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.scoreMap = scoreMap;
    }
    init() {
        this.dirCountMap.forEach(idMap => idMap.fill(0));
        this.conShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        this.gapShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        const { context, board, dirCountMap } = this;
        context.traverseAllDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const player = board[id];
            if (player >= 0) {
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
        });
        context.traverseLeftDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const player = board[id];
            if (player >= 0 && dirCountMap[dirType][id] === 1) {
                const revDirType = dirType ^ 1;
                this.updateConShapeCountMap(r, c, revDirType, 1);
            }
        });
        context.traverseLeftDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const revDirType = dirType ^ 1;
            if (board[id] < 0) {
                const [r0, c0] = context.move(r, c, revDirType, 1);
                const id0 = context.idxIfValid(r0, c0);
                if (id0 < 0 || board[id0] < 0)
                    return;
                const player = board[id0];
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                if (id2 < 0 || board[id2] !== player)
                    return;
                const lftCnt = dirCountMap[revDirType][id0];
                const rhtCnt = dirCountMap[dirType][id2];
                const [startR, startC] = context.move(r, c, revDirType, lftCnt);
                this.updateGapShapeCountMap(player, startR, startC, lftCnt, rhtCnt, dirType, 1);
            }
        });
    }
    beforeForward(r, c) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(r, c, dirType, -1);
        }
    }
    afterForward(r, c) {
        const { context, board, dirCountMap } = this;
        const id = context.idx(r, c);
        const player = board[id];
        {
            for (const dirType of gomokuDirectionTypes) {
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
            for (const dirType of gomokuDirectionTypes) {
                const [dr, dc] = gomokuDirections[dirType];
                const countMap = dirCountMap[dirType];
                const cnt = countMap[id];
                for (let r2 = r, c2 = c;;) {
                    r2 = r2 - dr;
                    c2 = c2 - dc;
                    if (context.isInvalidPos(r2, c2))
                        break;
                    const id2 = context.idx(r2, c2);
                    if (board[id2] !== player)
                        break;
                    countMap[id2] += cnt;
                }
            }
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(r, c, dirType, 1);
        }
    }
    beforeRollback(r, c) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(r, c, dirType, -1);
        }
    }
    afterRollback(r, c, player) {
        const { context, board, dirCountMap } = this;
        const id = context.idx(r, c);
        {
            for (const dirType of gomokuDirectionTypes) {
                const [dr, dc] = gomokuDirections[dirType];
                const countMap = dirCountMap[dirType];
                const cnt = countMap[id];
                for (let r2 = r, c2 = c;;) {
                    r2 = r2 - dr;
                    c2 = c2 - dc;
                    const id2 = context.idxIfValid(r2, c2);
                    if (id2 < 0 || board[id2] !== player)
                        break;
                    countMap[id2] -= cnt;
                }
            }
            for (const dirType of gomokuDirectionTypes)
                dirCountMap[dirType][id] = 0;
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(r, c, dirType, 1);
        }
    }
    score(currentPlayer, scoreForPlayer) {
        if (this.hasReachedTheLimit(scoreForPlayer))
            return Number.POSITIVE_INFINITY;
        if (this.hasReachedTheLimit(scoreForPlayer ^ 1))
            return Number.NEGATIVE_INFINITY;
        const { context, conShapeCountMap, gapShapeCountMap, scoreMap } = this;
        const conCountMap = conShapeCountMap[scoreForPlayer];
        const gapCountMap = gapShapeCountMap[scoreForPlayer];
        let score = 0;
        for (let cnt = 1; cnt < context.MAX_INLINE; ++cnt) {
            const [a, b, c] = conCountMap[cnt];
            const [x, y, z] = scoreMap.continuously[cnt];
            score += a * x + b * y + c * z;
        }
        for (let cnt = 1; cnt <= context.MAX_INLINE; ++cnt) {
            const [a, b, c] = gapCountMap[cnt];
            const [x, y, z] = scoreMap.gap[cnt];
            score += a * x + b * y + c * z;
        }
        if (currentPlayer === scoreForPlayer)
            return score;
        const buffer = Math.random() * context.NEXT_MOVER_BUFFER_FAC;
        return score + score * buffer;
    }
    hasReachedTheLimit(player) {
        const { context, conShapeCountMap } = this;
        return conShapeCountMap[player][context.MAX_INLINE].some(x => x > 0);
    }
    toJSON() {
        const { dirCountMap, conShapeCountMap, gapShapeCountMap } = this;
        return { dirCountMap, conShapeCountMap, gapShapeCountMap };
    }
    updateRelatedConShapeCountMap(centerR, centerC, dirType, v) {
        const { context, board, dirCountMap } = this;
        const THRESHOLD = context.MAX_INLINE * 2 + 1;
        const revDirType = dirType ^ 1;
        const [dr, dc] = gomokuDirections[dirType];
        let r = centerR - dr * context.MAX_INLINE;
        let c = centerC - dc * context.MAX_INLINE;
        let steps = 0;
        while (context.isInvalidPos(r, c)) {
            r += dr;
            c += dc;
            steps += 1;
        }
        while (steps < THRESHOLD) {
            const id = context.idxIfValid(r, c);
            if (id < 0)
                return;
            const player = board[id];
            if (player < 0) {
                r += dr;
                c += dc;
                steps += 1;
                continue;
            }
            const cnt = dirCountMap[dirType][id];
            if (steps === 0) {
                const rhtCnt = dirCountMap[dirType][id];
                const [r2, c2] = context.move(r, c, dirType, rhtCnt - 1);
                this.updateConShapeCountMap(r2, c2, revDirType, v);
            }
            else {
                this.updateConShapeCountMap(r, c, dirType, v);
            }
            r += dr * cnt;
            c += dc * cnt;
            steps += cnt;
        }
    }
    updateConShapeCountMap(startR, startC, dirType, v) {
        const id = this.context.idx(startR, startC);
        const player = this.board[id];
        const cnt = this.dirCountMap[dirType][id];
        const countOfFreeSide = this.countFreeSide(player, startR, startC, cnt, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.conShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    updateRelatedGapShapeCountMap(centerR, centerC, dirType, v) {
        const { context, board, dirCountMap } = this;
        const THRESHOLD = context.MAX_INLINE * 2 + 1;
        const revDirType = dirType ^ 1;
        const [dr, dc] = gomokuDirections[dirType];
        let r = centerR - dr * context.MAX_INLINE;
        let c = centerC - dc * context.MAX_INLINE;
        let steps = 0;
        for (; context.isInvalidPos(r, c); r += dr, c += dc)
            steps += 1;
        for (; steps < THRESHOLD; r += dr, c += dc, steps += 1) {
            const id = context.idxIfValid(r, c);
            if (id < 0)
                return;
            const player = board[id];
            if (player < 0)
                continue;
            const lftCnt = dirCountMap[revDirType][id];
            const rhtCnt = dirCountMap[dirType][id];
            if (steps < 2) {
                const [r2, c2] = context.move(r, c, dirType, rhtCnt - 1);
                this.detectGapShape(r2, c2, revDirType, v);
            }
            r -= dr * (lftCnt - 1);
            c -= dc * (lftCnt - 1);
            steps -= lftCnt - 1;
            break;
        }
        for (; steps < THRESHOLD;) {
            const id = context.idx(r, c);
            if (id < 0)
                return;
            const player = board[id];
            if (player < 0) {
                r += dr;
                c += dc;
                steps += 1;
                continue;
            }
            this.detectGapShape(r, c, dirType, v);
            const rhtCnt = dirCountMap[dirType][id];
            r += dr * rhtCnt;
            c += dc * rhtCnt;
            steps += rhtCnt;
        }
    }
    detectGapShape(r, c, dirType, v) {
        const { context, board, dirCountMap } = this;
        const id = context.idx(r, c);
        const lftCnt = dirCountMap[dirType][id];
        const [r1, c1] = context.move(r, c, dirType, lftCnt);
        const id1 = context.idxIfValid(r1, c1);
        if (id1 < 0 || board[id1] >= 0)
            return;
        const player = board[id];
        const [r2, c2] = context.move(r1, c1, dirType, 1);
        const id2 = context.idxIfValid(r2, c2);
        if (id2 < 0 || board[id2] !== player)
            return;
        const rhtCnt = dirCountMap[dirType][id2];
        this.updateGapShapeCountMap(player, r, c, lftCnt, rhtCnt, dirType, v);
    }
    updateGapShapeCountMap(player, startR, startC, lftCnt, rhtCnt, dirType, v) {
        const threshold = this.context.MAX_INLINE - 1;
        if (lftCnt >= threshold || rhtCnt >= threshold)
            return;
        const countOfFreeSide = this.countFreeSide(player, startR, startC, lftCnt + rhtCnt + 1, dirType);
        const normalizedCnt = Math.min(lftCnt + rhtCnt, this.context.MAX_INLINE);
        this.gapShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    countFreeSide(player, startR, startC, cnt, dirType) {
        const { context, board } = this;
        const { MAX_INLINE } = context;
        const revDirType = dirType ^ 1;
        const [r0, c0] = context.move(startR, startC, revDirType, 1);
        const id0 = context.idxIfValid(r0, c0);
        const isFreeSide0 = id0 >= 0 && board[id0] < 0 ? 1 : 0;
        const [r2, c2] = context.move(startR, startC, dirType, cnt);
        const id2 = context.idxIfValid(r2, c2);
        const isFreeSide2 = id2 >= 0 && board[id2] < 0 ? 1 : 0;
        const countOfFreeSide = isFreeSide0 + isFreeSide2;
        if (countOfFreeSide === 0)
            return 0;
        let maxPossibleCnt = cnt + countOfFreeSide;
        if (maxPossibleCnt < MAX_INLINE) {
            if (isFreeSide0) {
                const [dr, dc] = gomokuDirections[revDirType];
                for (let r = r0, c = c0; maxPossibleCnt < MAX_INLINE; ++maxPossibleCnt) {
                    r += dr;
                    c += dc;
                    const id = context.idxIfValid(r, c);
                    if (id < 0 || (board[id] >= 0 && board[id] !== player))
                        break;
                }
            }
            if (isFreeSide2) {
                const [dr, dc] = gomokuDirections[dirType];
                for (let r = r2, c = c2; maxPossibleCnt < MAX_INLINE; ++maxPossibleCnt) {
                    r += dr;
                    c += dc;
                    const id = context.idxIfValid(r, c);
                    if (id < 0 || (board[id] >= 0 && board[id] !== player))
                        break;
                }
            }
        }
        return maxPossibleCnt < MAX_INLINE ? 0 : countOfFreeSide;
    }
}

class GomokuState {
    context;
    countMap;
    candidateSet;
    board;
    placedCount;
    constructor(context, scoreMap) {
        this.context = context;
        this.board = new Int32Array(context.TOTAL_POS);
        this.countMap = new GomokuCountMap(context, this.board, scoreMap);
        this.candidateSet = new Set();
        this.placedCount = 0;
    }
    init(pieces = []) {
        const { context, board, candidateSet } = this;
        this.placedCount = pieces.length;
        board.fill(-1);
        candidateSet.clear();
        for (const { r, c, p } of pieces) {
            const id = context.idx(r, c);
            board[id] = p;
            candidateSet.delete(id);
            for (const [r2, c2] of context.validNeighbors(r, c)) {
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0)
                    candidateSet.add(id2);
            }
        }
        this.countMap.init();
    }
    forward(r, c, player) {
        const { context, board, candidateSet } = this;
        const id = context.idxIfValid(r, c);
        if (id < 0 || board[id] >= 0)
            return;
        this.beforeForward(r, c);
        {
            this.placedCount += 1;
            board[id] = player;
            candidateSet.delete(id);
            for (const [r2, c2] of context.validNeighbors(r, c)) {
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0)
                    candidateSet.add(id2);
            }
        }
        this.afterForward(r, c);
    }
    rollback(r, c) {
        const { context, board, candidateSet } = this;
        const id = context.idxIfValid(r, c);
        if (id < 0 || board[id] < 0)
            return;
        const player = board[id];
        this.beforeRollback(r, c);
        {
            this.placedCount -= 1;
            board[id] = -1;
            if (this.hasPlacedNeighbors(r, c))
                candidateSet.add(id);
            for (const [r2, c2] of context.validNeighbors(r, c)) {
                const id2 = context.idx(r2, c2);
                if (board[id2] >= 0 || !this.hasPlacedNeighbors(r2, c2)) {
                    candidateSet.delete(id2);
                }
            }
        }
        this.afterRollback(r, c, player);
    }
    expand(currentPlayer, scoreForPlayer) {
        const { context, board, candidateSet } = this;
        const candidates = [];
        for (const id of candidateSet) {
            if (board[id] >= 0)
                continue;
            const [r, c] = context.reIdx(id);
            candidates.push({ r, c, score: 0 });
        }
        for (const candidate of candidates) {
            const { r, c } = candidate;
            this.forward(r, c, currentPlayer);
            candidate.score = this.score(currentPlayer, scoreForPlayer);
            this.rollback(r, c);
        }
        return candidates;
    }
    score(currentPlayer, scoreForPlayer) {
        return (this.countMap.score(currentPlayer, scoreForPlayer) -
            this.countMap.score(currentPlayer, scoreForPlayer ^ 1));
    }
    isFinal() {
        const { context, countMap, placedCount } = this;
        if (placedCount === context.TOTAL_POS)
            return true;
        for (let player = 0; player < context.TOTAL_PLAYERS; ++player) {
            if (countMap.hasReachedTheLimit(player))
                return true;
        }
        return false;
    }
    randomMove() {
        const { context, board, candidateSet } = this;
        for (const id of candidateSet) {
            const [r, c] = context.reIdx(id);
            for (const [r2, c2] of context.validNeighbors(r, c)) {
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0)
                    return [r2, c2];
            }
        }
        return [-1, -1];
    }
    hasPlacedNeighbors(r, c) {
        const { context, board } = this;
        for (const [r2, c2] of context.validNeighbors(r, c)) {
            const id2 = context.idx(r2, c2);
            if (board[id2] >= 0)
                return true;
        }
        return false;
    }
    beforeForward(r, c) {
        this.countMap.beforeForward(r, c);
    }
    afterForward(r, c) {
        this.countMap.afterForward(r, c);
    }
    beforeRollback(r, c) {
        this.countMap.beforeRollback(r, c);
    }
    afterRollback(r, c, player) {
        this.countMap.afterRollback(r, c, player);
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

const createScoreMap = (MAX_INLINE) => {
    const continuously = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    const gap = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    let maxValue = 2 ** 30;
    const QUARTER_OF_MAX_VALUE = maxValue / 4;
    let cnt = MAX_INLINE;
    continuously[cnt] = [maxValue, maxValue, maxValue];
    gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    if (MAX_INLINE > 1) {
        cnt -= 1;
        continuously[cnt] = [0, QUARTER_OF_MAX_VALUE, maxValue / 2];
        gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    }
    maxValue /= 8;
    for (; cnt > 0; --cnt, maxValue /= 8) {
        continuously[cnt] = [0, maxValue / 4, maxValue / 2];
        gap[cnt] = [0, maxValue / 8, maxValue / 4];
    }
    return { continuously, gap };
};

class GomokuSolution {
    MAX_DEPTH;
    context;
    state;
    stateCompressor;
    stateCache;
    scoreForPlayer;
    bestR;
    bestC;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE = 5, MAX_DEPTH = 3, NEXT_MOVER_BUFFER_FAC, scoreMap) {
        const context = new GomokuContext(MAX_ROW, MAX_COL, MAX_INLINE, NEXT_MOVER_BUFFER_FAC);
        const _scoreMap = scoreMap ?? createScoreMap(context.MAX_INLINE);
        this.MAX_DEPTH = Math.max(1, Math.round(MAX_DEPTH));
        this.context = context;
        this.state = new GomokuState(context, _scoreMap);
        this.stateCompressor = new GomokuStateCompressor(BigInt(context.TOTAL_POS));
        this.stateCache = new Map();
        this.scoreForPlayer = -1;
        this.bestR = -1;
        this.bestC = -1;
    }
    init(pieces) {
        this.state.init(pieces);
        this.stateCache.clear();
    }
    move(r, c, p) {
        this.state.forward(r, c, p);
    }
    minmaxMatch(currentPlayer) {
        if (this.state.isFinal())
            return [-1, -1];
        this.stateCache.clear();
        this.scoreForPlayer = currentPlayer;
        this.bestR = this.bestC = -1;
        this.alphaBeta(currentPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, this.stateCompressor.INITIAL_STATE, 0);
        const { bestR: r, bestC: c } = this;
        return this.context.isValidPos(r, c) ? [r, c] : this.state.randomMove();
    }
    alphaBeta(player, alpha, beta, cur, prevState, stateScore) {
        const { context, state, stateCompressor, stateCache, scoreForPlayer } = this;
        if (cur === this.MAX_DEPTH || state.isFinal())
            return stateScore;
        const candidates = state.expand(player, scoreForPlayer);
        if (candidates.length <= 0)
            return stateScore;
        if (player === scoreForPlayer) {
            candidates.sort((x, y) => y.score - x.score);
        }
        else {
            candidates.sort((x, y) => x.score - y.score);
        }
        if (cur === 0 && this.bestR === -1) {
            const { r, c } = candidates[0];
            this.bestR = r;
            this.bestC = c;
        }
        for (const { r, c, score } of candidates) {
            const id = context.idx(r, c);
            const nextState = stateCompressor.compress(cur, prevState, BigInt(id));
            let gamma = stateCache.get(nextState);
            if (gamma === undefined) {
                state.forward(r, c, player);
                gamma = this.alphaBeta(player ^ 1, alpha, beta, cur + 1, nextState, score);
                state.rollback(r, c);
                stateCache.set(nextState, gamma);
            }
            if (cur === 0) {
                if (alpha < gamma) {
                    this.bestR = r;
                    this.bestC = c;
                }
            }
            if (player === scoreForPlayer) {
                if (alpha < gamma)
                    alpha = gamma;
            }
            else {
                if (beta > gamma)
                    beta = gamma;
            }
            if (beta <= alpha)
                break;
        }
        return player === scoreForPlayer ? alpha : beta;
    }
}

exports.GomokuContext = GomokuContext;
exports.GomokuCountMap = GomokuCountMap;
exports.GomokuSolution = GomokuSolution;
exports.GomokuState = GomokuState;
exports.GomokuStateCompressor = GomokuStateCompressor;
exports.createScoreMap = createScoreMap;
exports.gomokuDirectionTypes = gomokuDirectionTypes;
exports.gomokuDirections = gomokuDirections;
exports.leftHalfGomokuDirectionTypes = leftHalfGomokuDirectionTypes;
exports.rightHalfGomokuDirectionTypes = rightHalfGomokuDirectionTypes;
